<?php

declare(strict_types = 1);

namespace Drupal\Tests\phpexcel\Functional;

use Drupal\Component\Utility\Random;
use Drupal\node\Entity\Node;
use Drupal\Tests\BrowserTestBase;
use Drupal\phpexcel\PHPExcel;

/**
 * @file
 * Defines the test case for phpexcel.
 */
/**
 *
 */
class PHPExcelTest extends BrowserTestBase {

  protected $singleworksheetfile;
  protected $issue_1988868_file;
  protected $multiple_worksheet_file;
  protected $no_headers_file;
  protected $db_result_file;
  protected $template_file;
  protected $directory;

  protected $phpexcel;

  protected $fileSystem;

  /**
   * The entity storage for phpexcel config entities.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected $storage;
  /**
   * A user with administration permissions.
   *
   * @var \Drupal\user\UserInterface
   */
  protected $account;

  /**
   * Modules to enable.
   *
   * @var array
   */
  protected static $modules = ['node', 'phpexcel'];

  /**
   * Theme to enable.
   *
   * @var string
   */
  protected $defaultTheme = 'stark';

  /**
   *
   */
  public static function getInfo() {
    return [
      'name' => 'PHPExcel',
      'description' => 'Test the module API functions',
      'group' => 'PHPExcel',
    ];
  }

  /**
   *
   */
  protected function setUp(): void {
    parent::setUp();

    $this->phpexcel = \Drupal::service('phpexcel');

    $this->fileSystem = \Drupal::service('file_system');
    $defaultScheme = \Drupal::config('system.file')->get('default_scheme');
    $confPath = \Drupal::getContainer()->getParameter('site.path');

    $this->directory = \Drupal::state()->get('file_' . $defaultScheme . '_path', $confPath . '/files');

    // Make sure the directory is writable.
    $nonWritableMsg = $this->t("The @dir directory exists and is writable.", ['@dir' => $this->directory])->render();
    $this->assertDirectoryIsWritable($this->directory, $nonWritableMsg);

  }

  /**
   *
   */
  protected function tearDown(): void {
    // @unlink($this->singleworksheetfile);
    // @unlink($this->issue_1988868_file);
    // @unlink($this->multiple_worksheet_file);
    // @unlink($this->no_headers_file);
    // @unlink($this->db_result_file);
    // @unlink($this->template_file);
    parent::tearDown();
  }

  /**
   * Test a simple, single worksheet Excel export.
   */
  public function testSingleWorksheetExport() {
    /**
     * Export
     */
    // Prepare the data.
    $headers = ['Header 1', 'Header 2'];

    $data = [
      ['Data 1.1', 'Data 1.2'],
      ['Data 2.1', 'Data 2.2'],
      ['Data 3.1', 'Data 3.2'],
    ];

    // Create a file path
    // $correct_path = file_create_filename('phpexcel_test1.xls', $this->directory);.
    $correct_path = $this->fileSystem->createFilename('phpexcel_test1.xls', $this->directory);

    // The filename will be munged by the export function, so:
    $this->singleworksheetfile = $this->phpexcel->mungeFilename($correct_path);

    // Create a wrong path. Should not be able to export.
    $wrong_path = 'path/to/nowhere/file.xls';

    // Should fail.
    $this->assertEquals(
      PHPEXCEL::PHPEXCEL_ERROR_PATH_NOT_WRITABLE,
      $this->phpexcel->export($headers, $data, $wrong_path),
      'Passed an incorrect path'
    );

    // Should pass.
    $this->assertEquals(
      PHPEXCEL::PHPEXCEL_SUCCESS,
      $this->phpexcel->export($headers, $data, $correct_path),
      $this->t('Exported data to :path', [':path' => $this->singleworksheetfile])->render()
    );

    // Should pass.
    $this->assertTrue(filesize($this->singleworksheetfile) > 0, 'Filesize should be bigger than 0');

    /**
     * Import and check.
     * Import, not keyed by headers
     */
    $data = $this->phpexcel->import($this->singleworksheetfile, FALSE);

    // Should pass.
    $this->assertTrue(!!$data, 'Import succeeded');

    // Should have 4 rows (3 rows + headers)
    $count = !empty($data[0]) ? count($data[0]) : 0;
    $this->assertTrue($count === 4, $this->t(':count rows, expect 4.', [':count' => $count])->render());

    // Should only have 2 cells.
    $count = !empty($data[0][0]) ? count($data[0][0]) : 0;
    $this->assertTrue($count === 2, $this->t(':count cells, expect 2.', [':count' => $count])->render());

    // First row, 1st cell should be 'Header 1'.
    $header = !empty($data[0][0][0]) ? $data[0][0][0] : 'no';
    $this->assertTrue($header === 'Header 1', 'First row, 1st cell data should be "Header 1"');

    // Second row, 1st cell should be 'Data 1.1'.
    $header = !empty($data[0][1][0]) ? $data[0][1][0] : 'no';
    $this->assertTrue($header === 'Data 1.1', 'Second row, 1st cell data should be "Data 1.1"');
  }

  /**
   * A cell with a value of '0' must get exported as such.
   *
   * @see http://drupal.org/node/1988868
   */
  public function testIssue1988868() {
    /**
     * Export
     */
    // Prepare the data.
    $headers = ['Header 1', 'Header 2'];

    $data = [
      ['0', 'Data 1.2'],
      ['Data 2.1', '0'],
      ['0', '0'],
    ];

    // Create a file path.
    $correct_path = $this->fileSystem->createFilename('phpexcel_test1.xls', $this->directory);

    // The filename will be munged by the export function, so:
    $this->issue_1988868_file = $this->phpexcel->mungeFilename($correct_path);

    // Should pass.
    $this->assertEquals(
      PHPEXCEL::PHPEXCEL_SUCCESS,
      $this->phpexcel->export($headers, $data, $correct_path),
      $this->t('Exported data to :path', [':path' => $this->issue_1988868_file])->render()
    );

    // Should pass.
    $this->assertTrue(filesize($this->issue_1988868_file) > 0, 'Filesize should be bigger than 0');

    /**
     * Import and check.
     * Import, not keyed by headers
     */
    $data = $this->phpexcel->import($this->issue_1988868_file, FALSE);

    // Should pass.
    $this->assertTrue(!!$data, 'Import succeeded');

    // Second row, 1st cell should be 'Data 1.1'.
    $cell = $data[0][1][0];
    $this->assertTrue($cell === '0', 'Second row, 1st cell data should be "0"');

    // Second row, 2nd cell should be 'Data 1.1'.
    $cell = $data[0][1][1];
    $this->assertTrue($cell === 'Data 1.2', 'Second row, 2nd cell data should be "Data 1.2"');

    // Third row, 2nd cell should be '0'.
    $cell = $data[0][2][1];
    $this->assertTrue($cell === '0', 'Third row, 2nd cell data should be "0"');

    // Fourth row, 1st cell should be '0'.
    $cell = $data[0][3][0];
    $this->assertTrue($cell === '0', 'Fourth row, 1st cell data should be "0"');

    // Fourth row, 2nd cell should be '0'.
    $cell = $data[0][3][1];
    $this->assertTrue($cell === '0', 'Fourth row, 2nd cell data should be "0"');
  }

  /**
   * Test multiple worksheet Excel export.
   */
  public function testMultipleWorksheetExport() {
    /**
     * Export
     */
    // Prepare data.
    $headers = [
      'Sheet 1' => ['Header 1.1', 'Header 1.2'],
      'Sheet 2' => ['Header 2.1', 'Header 2.2'],
    ];

    $data = [
      'Sheet 1' => [
        ['Data 1.1.1', 'Data 1.1.2'],
        ['Data 1.2.1', 'Data 1.2.2'],
        ['Data 1.3.1', 'Data 1.3.2'],
      ],
      'Sheet 2' => [
        ['Data 2.1.1', 'Data 2.1.2'],
        ['Data 2.2.1', 'Data 2.2.2'],
        ['Data 2.3.1', 'Data 2.3.2'],
      ],
    ];

    // Create a file path.
    $correct_path = $this->fileSystem->createFilename('phpexcel_test2.xls', $this->directory);

    // The filename will be munged by the export function, so:
    $this->multiple_worksheet_file = $this->phpexcel->mungeFilename($correct_path);

    // Should pass.
    $this->assertEquals(
      PHPEXCEL::PHPEXCEL_SUCCESS,
      $this->phpexcel->export($headers, $data, $correct_path),
      $this->t('Exported data to :path', [':path' => $this->multiple_worksheet_file])->render()
    );

    // Should pass.
    $this->assertTrue(filesize($this->multiple_worksheet_file) > 0, 'Filesize should be bigger than 0');

    /**
     * Import and check.
     * Import, keyed by headers
     */
    $data = $this->phpexcel->import($this->multiple_worksheet_file);

    // Should pass.
    $this->assertTrue(!!$data, 'Import succeeded');

    // Should have 3 rows.
    $count = !empty($data[0]) ? count($data[0]) : 0;
    $this->assertTrue($count === 3, $this->t(':count rows, expect 3', [':count' => $count])->render());

    // Should only have 2 cells.
    $count = !empty($data[0][0]) ? count($data[0][0]) : 0;
    $this->assertTrue($count === 2, $this->t(':count cells, expect 2', [':count' => $count])->render());

    // Should be keyed by headers.
    $this->assertTrue(isset($data[0][0]['Header 1.1']), 'Keyed by header ("Header 1.1")');
    $this->assertTrue(isset($data[1][0]['Header 2.2']), 'Keyed by header ("Header 2.2")');

    $header = !empty($data[0][0]['Header 1.1']) ? $data[0][0]['Header 1.1'] : 'no';
    $this->assertTrue($header === 'Data 1.1.1', 'Should be "Data 1.1.1"');

    $header = !empty($data[1][1]['Header 2.2']) ? $data[1][1]['Header 2.2'] : 'no';
    $this->assertTrue($header === 'Data 2.2.2', 'Should be "Data 2.2.2"');

    /**
     * Import and check.
     * Import with worksheet names.
     */
    $data = $this->phpexcel->import($this->multiple_worksheet_file, TRUE, TRUE);

    // Should pass.
    $this->assertTrue(!!$data, 'Import succeeded');

    // Should have 3 rows.
    $count = !empty($data['Sheet 1']) ? count($data['Sheet 1']) : 0;
    $this->assertTrue($count === 3, $this->t(':count rows, expect 3', [':count' => $count])->render());

    // Should be keyed by Worksheet name.
    $this->assertTrue(isset($data['Sheet 1']), $this->t('Imported with Worksheet names.')->render());
    $this->assertTrue(isset($data['Sheet 2']), $this->t('Imported with Worksheet names.')->render());

    // Should only have 2 cells.
    $count = !empty($data['Sheet 1'][0]) ? count($data['Sheet 1'][0]) : 0;
    $this->assertTrue($count === 2, $this->t(':count cells, expect 2', [':count' => $count])->render());

    // Should be keyed by headers.
    $this->assertTrue(isset($data['Sheet 1'][0]['Header 1.1']), 'Keyed by header ("Header 1.1")');
    $this->assertTrue(isset($data['Sheet 2'][0]['Header 2.2']), 'Keyed by header ("Header 2.2")');

    $header = !empty($data['Sheet 1'][0]['Header 1.1']) ? $data['Sheet 1'][0]['Header 1.1'] : 'no';
    $this->assertTrue($header === 'Data 1.1.1', 'Should be "Data 1.1.1"');

    $header = !empty($data['Sheet 2'][1]['Header 2.2']) ? $data['Sheet 2'][1]['Header 2.2'] : 'no';
    $this->assertTrue($header === 'Data 2.2.2', 'Should be "Data 2.2.2"');
  }

  /**
   * Test "ignore_headers" option.
   */
  public function testIgnoreHeaders() {
    /**
     * Export
     */
    // Prepare data.
    $data = [
      0 => [
        ['Data 1.1.1', 'Data 1.1.2'],
        ['Data 1.2.1', 'Data 1.2.2'],
        ['Data 1.3.1', 'Data 1.3.2'],
      ],
      'Sheet 2' => [
        ['Data 2.1.1', 'Data 2.1.2'],
        ['Data 2.2.1', 'Data 2.2.2'],
        ['Data 2.3.1', 'Data 2.3.2'],
      ],
    ];

    // Create a file path.
    $correct_path = $this->fileSystem->createFilename('phpexcel_test3.xls', $this->directory);

    // The filename will be munged by the export function, so:
    $this->no_headers_file = $this->phpexcel->mungeFilename($correct_path);

    // Should pass.
    $this->assertEquals(
      PHPEXCEL::PHPEXCEL_SUCCESS,
      $this->phpexcel->export(NULL, $data, $correct_path, ['ignore_headers' => TRUE]),
      $this->t('Exported data to :path', [':path' => $this->no_headers_file])->render()
    );

    // Should pass.
    $this->assertTrue(filesize($this->no_headers_file) > 0, 'Filesize should be bigger than 0');

    /**
     * Import and check.
     * Import, not keyed by headers
     */
    $data = $this->phpexcel->import($this->no_headers_file, FALSE);

    // Should pass.
    $this->assertTrue(!!$data, 'Import succeeded');

    // Should have 3 rows.
    $count = !empty($data[0]) ? count($data[0]) : 0;
    $this->assertTrue($count === 3, $this->t(':count rows, expect 3', [':count' => $count])->render());

    // Should only have 2 cells.
    $count = !empty($data[0][0]) ? count($data[0][0]) : 0;
    $this->assertTrue($count === 2, $this->t(':count cells, expect 2', [':count' => $count])->render());

    $header = !empty($data[0][0][0]) ? $data[0][0][0] : 'no';
    $this->assertTrue($header === 'Data 1.1.1', 'Should be "Data 1.1.1"');

    $header = !empty($data[1][1][1]) ? $data[1][1][1] : 'no';
    $this->assertTrue($header === 'Data 2.2.2', 'Should be "Data 2.2.2"');
  }

  /**
   * Test a simple, single worksheet Excel export.
   */
  public function testTemplateExport() {
    /**
     * Export
     */
    // Prepare the data.
    $data = [
      ['Data 1.1', 'Data 1.2'],
      ['Data 2.1', 'Data 2.2'],
      ['Data 3.1', 'Data 3.2'],
    ];

    // Options.
    $options = [
      'format' => 'xlsx',
      'template' => \Drupal::service('extension.list.module')->getPath('phpexcel') . '/tests/Functional/data/phpexcel.test.template.xlsx',
      'ignore_headers' => TRUE,
    ];

    // Create a file path.
    $correct_path = $this->fileSystem->createFilename('phpexcel_test4.xlsx', $this->directory);

    // The filename will be munged by the export function, so:
    $this->template_file = $this->phpexcel->mungeFilename($correct_path);

    // Should pass.
    $this->assertEquals(
      PHPEXCEL::PHPEXCEL_SUCCESS,
      $this->phpexcel->export([], $data, $correct_path, $options),
      $this->t('Exported data to :path', [':path' => $this->template_file])->render()
    );

    // Should pass.
    $this->assertTrue(filesize($this->template_file) > 0, 'Filesize should be bigger than 0');

    /**
     * Import and check.
     */
    $data = $this->phpexcel->import($this->template_file, FALSE);

    // Should pass.
    $this->assertTrue(!!$data, 'Import succeeded');

    // Should have 3 rows (3 rows and no headers)
    $count = !empty($data[0]) ? count($data[0]) : 0;
    $this->assertTrue($count === 3, $this->t(':count rows, expect 3', [':count' => $count])->render());

    // First row, 1st cell should be 'Data 1.1'.
    $header = !empty($data[0][0][0]) ? $data[0][0][0] : 'no';
    $this->assertTrue($header === 'Data 1.1', 'First row, 1st cell data should be "Data 1.1", and is ' . $header);

    // Second row, 1st cell should be 'Data 2.1'.
    $header = !empty($data[0][1][0]) ? $data[0][1][0] : 'no';
    $this->assertTrue($header === 'Data 2.1', 'Second row, 1st cell data should be "Data 2.1", and is ' . $header);
  }

  /**
   * Test db_result export.
   */
  public function testDBResultExport() {
    /**
     * Export
     */
    // Create 10 nodes.
    for ($i = 10; $i > 0; $i--) {
      $this->_createNode();
    }

    // Get the db query result.
    $result = \Drupal::database()->select('node_field_data', 'n')
      ->fields('n', ['nid', 'title'])
      ->execute();

    // Create a path.
    $correct_path = $this->fileSystem->createFilename('phpexcel_test4.xlsx', $this->directory);

    // The filename will be munged by the export function, so:
    $this->db_result_file = $this->phpexcel->mungeFilename($correct_path);

    // // Try exporting to Excel 2007.
    // $options = [
    //   'format' => 'xlsx',
    //   'creator' => 'SimpleTest',
    //   'title' => 'DBResult',
    //   'subject' => 'test',
    //   'description' => 'my description',
    // ];
    // Should pass.
    $this->assertEquals(
      PHPEXCEL::PHPEXCEL_SUCCESS,
      $this->phpexcel->exportDBResult($result, $correct_path),
      $this->t('Exported data to :path', [':path' => $this->db_result_file])->render()
    );

    // Should pass.
    $this->assertTrue(filesize($this->db_result_file) > 0, 'Filesize should be bigger than 0');

    /**
     * Import and check.
     * Import, cells keyed by headers.
     */
    $data = $this->phpexcel->import($this->db_result_file);

    // Should pass.
    $this->assertTrue(!!$data, 'Import succeeded');

    // Should have 10 rows.
    $count = count($data[0]);
    $this->assertTrue($count === 10, $this->t(':count rows, expect 10', [':count' => $count])->render());

    // Should only have 2 cells.
    $count = count($data[0][0]);
    $this->assertTrue($count === 2, $this->t(':count cells, expect 2', [':count' => $count])->render());

    // Should be keyed by headers (nid & title)
    $this->assertTrue(isset($data[0][0]['nid']), 'Keyed by header (nid)');
    $this->assertTrue(isset($data[0][1]['title']), 'Keyed by header (title)');
  }

  /**
   * Test the ability to pass cutom methods and arguments on import to the Reader.
   */
  public function testCustomMethodCalls() {
    /**
     * Export
     */
    // Prepare data.
    $headers = [
      'Sheet 1' => ['Header 1.1', 'Header 1.2'],
      'Sheet 2' => ['Header 2.1', 'Header 2.2'],
    ];

    $data = [
      'Sheet 1' => [
        ['Data 1.1.1', 'Data 1.1.2'],
        ['Data 1.2.1', 'Data 1.2.2'],
        ['Data 1.3.1', 'Data 1.3.2'],
      ],
      'Sheet 2' => [
        ['Data 2.1.1', 'Data 2.1.2'],
        ['Data 2.2.1', 'Data 2.2.2'],
        ['Data 2.3.1', 'Data 2.3.2'],
      ],
    ];

    // Create a file path.
    $correct_path = $this->fileSystem->createFilename('phpexcel_test2.xls', $this->directory);

    // The filename will be munged by the export function, so:
    $this->multiple_worksheet_file = $this->phpexcel->mungeFilename($correct_path);

    // Should pass.
    $this->assertEquals(
      PHPEXCEL::PHPEXCEL_SUCCESS,
      $this->phpexcel->export($headers, $data, $correct_path),
      $this->t('Exported data to :path', [':path' => $this->multiple_worksheet_file])->render()
    );

    // Should pass.
    $this->assertTrue(filesize($this->multiple_worksheet_file) > 0, 'Filesize should be bigger than 0');

    /**
     * Import and check.
     * Pass a method call to only import a specific worksheet.
     */
    $data = $this->phpexcel->import($this->multiple_worksheet_file, TRUE, TRUE, ['setLoadSheetsOnly' => ['Sheet 1']]);

    // Should pass.
    $this->assertTrue(!!$data, 'Import succeeded');

    // Should have 3 rows.
    $count = !empty($data['Sheet 1']) ? count($data['Sheet 1']) : 0;
    $this->assertTrue($count === 3, $this->t(':count rows, expect 3', [':count' => $count])->render());

    // Should be empty.
    $count = !empty($data['Sheet 2']) ? count($data['Sheet 2']) : 0;
    $this->assertTrue($count === 0, $this->t(':count rows, expect 0', [':count' => $count])->render());
  }

  /**
   *
   */
  protected function _createNode() {
    $random = new Random();

    $node = Node::create(['type' => 'page']);
    $node->title = $random->name(8);
    $node->body = $random->name(20);
    $node->save();
  }

}
